from dnalib.core import *
from dnalib.custom import *
from dnalib.log import *
from dnalib.utils import *
from dnalib.tables import *

#  
#   Description:
#       This method creates a table based on our patterns (field comments, table comments, view for hashing and etc).
#
#   Parameters:      
#       schema[df.schema] = table schema (same of df.schema)
#       layer[str] = table layer (bronze, silver, gold, diamond or export)
#       table_name[str] = the name of the table.
#       yml[yml object, string or None] = yml file path (it is not required to create a bronze table). Default is none.
#
def create_table(schema, layer, table_name, yml={}, partition_fields=[], anonimized_fields=[], comment={}, comment_squad={}, fields={}, tbl_properties={}, replace=False):
    log(__name__).info(f"Running create_table method for {layer}.{table_name}")
    if layer != "bronze":
        ct = CreateTableSilver(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, tbl_properties, replace).execute()        
        CreateView(schema, layer, table_name, ct.yml, ct.tbl_comment.parsed_comment, anonimized_fields, comment, comment_squad, fields, replace).execute()
    else:
        CreateTableBronze(schema, table_name, yml, partition_fields, tbl_properties, replace).execute()
    log(__name__).info(f"End create_table method for {layer}.{table_name}")

#  
#   Description:
#       This method creates a table in bronze layer based on our patterns (field comments, table comments, view for hashing and etc).
#
#   Parameters:      
#       schema[df.schema] = table schema (same of df.schema)
#       table_name[str] = the name of the table.
#       yml[yml object, string or None] = yml file path (it is not required to create a bronze table). Default is none.
#
def create_table_bronze(schema, table_name, yml={}, partition_fields=[], tbl_properties={}, replace=False):
    log(__name__).info(f"Running create_table_bronze method for bronze.{table_name}")
    CreateTableBronze(schema, table_name, yml, partition_fields, tbl_properties, replace).execute()
    log(__name__).info(f"End create_table_bronze method for bronze.{table_name}")

#  
#   Description:
#       This method creates a table in silver layer based on our patterns (field comments, table comments, view for hashing and etc).
#
#   Parameters:      
#       schema[df.schema] = table schema (same of df.schema)
#       table_name[str] = the name of the table.
#       yml[yml object, string or None] = yml file path (it is not required to create a bronze table). Default is none.
#
def create_table_silver(schema, table_name, yml={}, partition_fields=[], anonimized_fields=[], comment={}, comment_squad={}, fields={}, tbl_properties={}, replace=False):
    log(__name__).info(f"Running create_table_silver method for silver.{table_name}")
    ct = CreateTableSilver(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, tbl_properties, replace).execute()        
    CreateView(schema, layer, table_name, ct.yml, ct.tbl_comment.parsed_comment, anonimized_fields, comment, comment_squad, fields, replace).execute()
    log(__name__).info(f"End create_table_silver method for silver.{table_name}")

#  
#   Description:
#       This method drops a table and all its files in storage.
#
#   Parameters:      
#       layer[str] = table layer (bronze, silver, gold, diamond or export)
#       table_name[str] = the name of the table.
#
def drop_table(layer, table_name):
    log(__name__).info(f"Running drop_table method for {layer}.{table_name}")
    TableUtils.drop_table(layer, table_name)
    log(__name__).info(f"End drop_table method for {layer}.{table_name}")

#  
#   Description:
#       This method drops a view from a table.
#
#   Parameters:      
#       layer[str] = table layer (bronze, silver, gold, diamond or export)
#       table_name[str] = the name of the table.
#
def drop_view(layer, table_name):
    log(__name__).info(f"Running drop_view method for {layer}.{table_name}")
    TableUtils.drop_view(layer, table_name)
    log(__name__).info(f"End drop_view method for {layer}.{table_name}")    


from pyspark.sql.functions import from_utc_timestamp, current_timestamp, col, upper, trim, regexp_replace, to_date, when, date_format, lit, coalesce, xxhash64, expr
from dnalib.writer import WriteModes, BatchTableWriter
from dnalib.utils import Utils
from dnalib.log import log

class LayerTable:    
    """ 
        A classe LayerTable é uma implementação de alto nível de uma tabela baseado na arquitetura que seguimos. Deve ser usada como base para implementar tabelas específicas (silver, gold, diamond e etc). A ideia geral é sempre carregar de um dataframe source (camada de entrada) para um target (camada de saída).
    
        Args:
            layer (str): parâmetro que indica a layer de destino onde o target_df será persistido.
            table_name (str): string que representa o nome da tabela.            
            source_df (spark Dataframe, optional): dataframe do spark onde os dados serão consumidos para compor o target_df. 
            include_checksum (bool, optional): se verdadeiro adiciona o campo de checksum no target_df.                  

        Attributes:
            table_name (str): string que representa o nome da tabela.    
            source_df (spark Dataframe): dataframe do spark onde os dados serão consumidos para compor o target_df.                    
            include_checksum (bool): se verdadeiro adiciona o campo de checksum no target_df.     
            writer (BatchTableWriter): intância interna do BatchTableWriter para persistir o dataframe target_df.
            target_df (spark Dataframe): dataframe para onde os dados consumidos de source_df serão materializados. Deve ser populado pela implementação.
            has_add_checksum (bool): parâmetro interno para controle de se o dataframe target_df foi adicionado o parâmetro de checksum.
            has_add_data_carga (bool): parâmetro interno para controle de se o dataframe target_df foi adicionado o parâmetro de dataCarga.
            parsed_source_df_query (str): resultado do método parse_df_source_query() populado após a chamada do load_source_df_data().
            has_loaded (bool): parâmetro interno para controle de se o método load() foi chamado.
    """

    def __init__(self, layer, table_name, source_df=None, include_checksum=False):                
        """ Construtor da classe """
        self.table_name = table_name
        self.source_df = source_df 
        self.include_checksum = include_checksum
        self.writer = BatchTableWriter(self.table_name, layer)
        self.target_df = None      
        self.has_add_checksum = False  
        self.has_add_data_carga = False        
        self.has_loaded = False                   

    def parse_df_source(self):
        """
            Método abstrato para ser implementado. Deve retornar um dataframe que representa os dados da camada de entrada. 

            Returns:
                source_df (spark Dataframe): dataframe do spark onde os dados serão consumidos para compor o target_df.      
        """
        raise NotImplementedError("This method must be implemented")

    def load_fields_from_source(self):
        """
            Método abstrato para ser implementado. Deve retornar uma lista de fields que será populada no target_df. 

            Returns:
                list_of_fields (list): lista com as fields para gerar o dataframe target_df a partir do source_df.
        """
        raise NotImplementedError("This method must be implemented")

    def load_source_df_data(self):
        """
            Método que faz a chamada do parser do dataframe gerando o dataframe de entrada (source_df). O source_df é preenchido nesse processo. Internamente usa a implementação do parse_df_source().
        """
        # assert that the dataframe is not None
        if self.source_df is None:                      
            self.source_df = self.parse_df_source()  

    def load_target_df_data(self): 
        """
            Método que carrega as fields a partir do dataframe source. Internamente usa a implementação de load_fields_from_source().
        """       
        # get fields from source
        self.target_df = self.source_df.select(self.load_fields_from_source())

    def add_checksum(self):       
        """ 
            Método que adiciona ao final do dataframe o parâmetro de checksum no dataframe target_df usando a função xxhash64. Por default só é possível usar campos primitivos para gerar o hash (string, int, float e etc). 
        """         
        columns = self.target_df.columns
        # it prevents to add checksum with dataCarga parameter   
        if "dataCarga" in columns:         
            columns.remove("dataCarga")        
        # It appends checkum as the last parameter
        self.target_df = self.target_df.select("*", xxhash64(*columns).alias("checksum"))
        # boolean parameter to internal control
        self.has_add_checksum = True  
    
    def add_data_carga(self):
        """ 
            Método que adiciona a dataCarga no começo do dataframe, isso garante que o parâmetro de dataCarga pode ser usado no z-order. 
        """
        # It guarantees that dataCarga is the first parameter
        self.target_df = self.target_df.select(from_utc_timestamp(current_timestamp(), "America/Sao_Paulo").alias("dataCarga"), "*")
        # boolean parameter to internal control
        self.has_add_data_carga = True  

    def load(self):        
        """
            Método que executa os steps pra carregar o dataframe que representa os dados para o target_df. Primeiro são carregados os dados do source para o target (usando a estratégia implementada no método load_source_df_data()). Em seguida, executa as tratativas dos campos que são levados de uma camada para a outra. Finalmente, os campos de checksum (se include_checksum=True) e dataCarga são adicionados.

            Returns:
                self: a instância da classe LayerTable.
        """
        # load inicial dataframe
        self.load_source_df_data()

        # load fields from source layer
        self.load_target_df_data()

        # add checksum column at the target dataframe
        if self.include_checksum:
            self.add_checksum()

        # automatic append dataCarga
        self.add_data_carga()

        # mark method as loaded
        self.has_loaded = True        
        log(__name__).info(f"Sucessfully loaded {self.target_df.count()} rows.") 
        return self

    def persist(self, mode=WriteModes.OVERWRITE, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None):             
        """ 
            Método wrapper do BatchTableWriter.persist, automaticamente detecta se no dataframe foi adicionado o checksum e preeenche o campo has_checksum_field. 
        """
        self.writer.persist(self.target_df, mode=mode, partition_fields=partition_fields, optimize=optimize, merge_condition=merge_condition, has_checksum_field=self.has_add_checksum)      
        return self  


class ProcessLayerTable(LayerTable):
    """ 
        A classe ProcessLayerTable é uma implementação mais especializada da classe Table que implementa métodos para tratativa de campos (renonemar, formatar, fazer cast, etc).

        Args:
            layer (str): parâmetro que indica a layer de destino onde o target_df será persistido (passado para a classe pai internamente).
            table_name (str): string que representa o nome da tabela.            
            source_df (spark Dataframe, optional): dataframe do spark onde os dados serão consumidos para compor o target_df.    
            include_checksum (bool, optional): se verdadeiro adiciona o campo de checksum no target_df.         

        Attributes:
            real_types (list): lista de tipos reais válidos para cast.            
            valid_date_formats (list): lista de formatos válidos de converção inferida de datas.
    """

    real_types = ["float", "decimal", "double", "real"]
    valid_date_formats = ['M/d/yyyy', 'yyyy/M/d', 'd-M-yyyy', 'yyyy-M-d', 'yyyyMMdd', 'ddMMyyyy'] 

    def __init__(self, layer, table_name, source_df=None, include_checksum=False):
        """ Construtor da classe """
        super().__init__(layer, table_name, source_df, include_checksum)    
    
    def format_field(self, field, source_field_type, target_field_type, format_field):
        """
            Método que formata campos strings para tipos primitivos (string para float, decimal, double, real e date)

                .. code-block:: portugol
                    se format_field é verdadeiro:
                        se source é uma string então:
                            remova espaços e deixa o conteúdo do texto em maiúsculo.
                            se cast para float, decimal, real ou double então:
                                remove caracteres especiais (ex: R$:123.56 vai ser formatado para 123.56 antes do cast)
                            se cast para date:
                                tenta converter para qualquer um dos formatos de data válidos (M/d/yyyy, yyyy/M/d, d-M-yyyy, yyyy-M-d, yyyyMMdd, ddMMyyyy)
                    se format_field é string:
                        aplica a função expr para formatar o campo.

            Args:
                field (col): campo a ser formatado, um col para spark
                source_field_type (str): tipo do campo na camada de origem (ex: bronze)               
                target_field_type (str): tipo do campo na camada de destino (ex: silver, vai ser forçado cast após a formatação)
                format_field (bool or str): se True, indica que o campo deve ser formatado de acordo com seu tipo na camada de destino. Se str é executado via função expr.

            Returns:
                formated_field (col): campo formatado.
        """
        formated_field = field
        if format_field:
            # formata o campo usando a função expr do spark
            if isinstance(format_field, str):
                formated_field = expr(format_field)
            else:
                # format string fields from source         
                if source_field_type.lower() == "string":
                    formated_field = upper(trim(formated_field))                    

                    # format real types before cast
                    if target_field_type != None:
                        if any(real in target_field_type.lower() for real in self.real_types):
                            # remove non valid characters before cast
                            formated_field = regexp_replace(formated_field, '[^\d.,+-]?', '')

                        # try to infer date
                        if target_field_type.lower() == "date":             
                            list_of_formated_fields = [to_date(formated_field, date_to_format) for date_to_format in self.valid_date_formats]
                            formated_field = coalesce(*list_of_formated_fields)                                                  

        return formated_field      

    def rename_field(self, field, source_field_name, target_field_name=None):  
        """
            Esse método implementa um rename de maneira "safe", isto é, caso o campo tenha feito algum processamento anterior, garante que o nome seja mantido.

            Args:
                field (col): o campo que deve ser renomeado.
                source_field_name (str): string com o nome do campo na camada de origem.
                target_field_name (str or None): string com o nome do campo na camada de destino.
            
            Returns:
                field (col): o campo renomeado
        """      
        if target_field_name != None:
            return field.alias(target_field_name)
        else:
            return field.alias(source_field_name)
        
    def cast_field(self, field, field_type=None):      
        """
            Método que faz o cast de um campo para o tipo informado.

            Args:                
                field (col): o campo que deve ser formatado.                
                field_type (str): string com o tipo do campo (qualquer um da documentação [https://spark.apache.org/docs/latest/sql-ref-datatypes.html]).            
            Returns:                
                field (col): o campo após o cast.
        """  
        if field_type != None:
            return field.cast(field_type)
        else: 
            return field




import logging

def log(logger_name):   
    """
        Método simples que cria um logger para ser usado em todos os processos internos da dnalib. Garante que um handler é adicionado apenas uma vez ao logger_name.

        Args:
            logger_name (str): geralmente é passado o parâmetro __name__ do módulo, que permite uma identificação mais interessante nos arquivos de log.

        Returns:
            logger: objeto de logger.
    """ 
    logger = logging.getLogger(logger_name)
    if not logger.handlers:
        # set logger level
        logger.setLevel(logging.INFO)            
        logger.propagate = False

        # define handler and formatter
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")

        # add formatter to handler
        handler.setFormatter(formatter)

        # add handler to logger
        logger.addHandler(handler)
    return logger


from dnalib.log import log
from datetime import datetime
import pytz
import time
import pysftp


class Sftp:

    def __init__(self, username, password, host, port, max_retry=10, max_seconds_retry_delay=10):
        """ """
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.max_retry = max_retry
        self.max_seconds_retry_delay = max_seconds_retry_delay    
        self.connection = None

    def connect(self):             
        """ """
        # Set hostkeys to None to disable host key checking   
        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None
        # Count number of retries
        current_retry = 0
        currently_error = ""
        for retry in range(0, self.max_retry):            
            try:
                self.connection = pysftp.Connection(host=self.host, username=self.username, password=self.password, port=self.port, cnopts=cnopts)
                break
            except Exception as e:
                time.sleep(self.max_seconds_retry_delay)
                currently_error = e
                current_retry = current_retry + 1
        # If max retries is reached, raise an exception
        if current_retry == self.max_retry:
            log(__name__).error(f"You reached max number of retry, you must have some problem in connection:\n {currently_error}")            
            raise Exception(f"You reached max number of retry, you must have some problem in connection:\n {currently_error}")
        # if not happn we are sucessfully connected        
        log(__name__).info(f"Sucessfully connected to {self.host} with user {self.username}.")
        return self

    def listdir(self, remote_path):
        """ """
        for directory in self.connection.listdir(remote_path):
            yield directory

    def listdir_attr(self, remote_path):
        """ """
        for attr in self.connection.listdir_attr(remote_path):
            yield attr

    def disconnect(self):
        """ """
        self.connection.close()
        log(__name__).info(f"Disconnected from {self.host}.")    
        return self
    
    def add_watermark(self, remote_path):
        tz = pytz.timezone('America/Sao_Paulo')
        watermark = datetime.now(tz).strftime('%Y%m%d%H%M%S') 
        file_split = remote_path.rsplit(".", 1)
        return f"{file_split[0]}_{watermark}.{file_split[1]}"
    
    def remove(self, *files):
        """  """
        for file in files:
            try:
                self.connection.remove(file)
            except Exception as e:
                log(__name__).warning(f"Unable to remove file {file}:\n {e}")
        return self

    def upload(self, source_local_path, remote_path, include_watermark=True):
        """ """
        if include_watermark:
            remote_path = self.add_watermark(remote_path)
        try:
            # upload file to SFTP
            log(__name__).info(f"Uploading {source_local_path} to {remote_path}")
            self.connection.put(source_local_path, remote_path)            
        except Exception as e:
            log(__name__).error(f"Unable to upload file {source_local_path} to {remote_path}:\n {e}")
            raise Exception(f"Unable to upload file {source_local_path} to {remote_path}:\n {e}")
        log(__name__).info(f"Sucessfully uploaded {source_local_path} to {remote_path}")
        return self
    
class MftSftp(Sftp):

    def __init__(self, username, password, **kwargs):
        super().__init__(username, password, "mft.portoseguro.brasil", 2022, **kwargs)
        


from pyspark.sql import SparkSession
from dnalib.utils import TableUtils, Utils
from dnalib.log import log
from .tablecomment import TableComment

class CreateTable:
    """
        Classe de alto nível para o create table. Define o método abstrato de parser para as fields. Não deve ser alocada.

        Args:      
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or {}): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é {}.
            partition_fields (list, optional): lista de campos utilizados como partição da tabela.      
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou {}): instância do objeto yml, é {} caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            parsed_template (str): template da criação da tabela compilado, preenchido após a chamada do método template.
    """

    def __init__(self, 
                 schema,  
                 table_name, 
                 yml={},                  
                 partition_fields=[],
                 tbl_properties={},
                 replace=False):        
        self.table_name = table_name.strip().lower()
        self.schema = schema
        self.table_path = Utils.lakehouse_path(self.layer, self.table_name) 
        self.yml = yml
        if isinstance(self.yml, str):
            self.yml = Utils.safe_load_and_parse_yml(yml)
        self.partition_fields = self.yml.get("partition_fields", partition_fields)
        self.tbl_properties = self.yml.get("tbl_properties", tbl_properties)
        self.replace = replace         
        self.has_created_table = False         
        self.parsed_template = ""

    def parse_fields(self):
        """
            Método abstrato para ser implementado, faz o trabalho de parsear as fields para gerar o template do Create Table.
        """
        return NotImplementedError("Method parse_fields must be implemented")
    
    def template(self):        
        """
            Método que gera o template do Create Table e injeta no objeto self.parsed_template.
            
            .. code-block:: SQL

                CREATE OR REPLACE TABLE layer.table (
                    field_name field_type COMMENT '' nullable or not,
                    ...
                )
                USING DELTA
                TBLPROPERTIES (tbl_properties)
                PARTITIONED BY (partition_fields)

           Returns:
                self: a instância da classe CreateTable.
        """        
        # parse fields to string
        self.parse_fields()        
        # parse table partition_fields to a string
        partition_fields = ""
        if len(self.partition_fields) > 0:
            partition_fields = ", ".join(self.partition_fields)
            partition_fields = f"PARTITIONED BY ({partition_fields})"
        # tbl properties
        tbl_properties = ""
        if len(self.tbl_properties) > 0:
            parsed_tbl_properties = ", ".join([f"{key}='{value}'" for key, value in self.tbl_properties.items()])
            tbl_properties = f"TBLPROPERTIES ({parsed_tbl_properties})"
        # generate template for create table
        self.parsed_template = """
            CREATE OR REPLACE TABLE {}.{} (
                {}
            )
            USING delta
            {}
            {}
            LOCATION '{}'
        """.format(self.layer, self.table_name, self.parsed_fields, tbl_properties, partition_fields, self.table_path)
        return self    

    def execute(self):  
        """
            Método que executa o create table com base no template gerado. Caso a tabela já exista, não é executado nada a não ser que o parâmetro replace=True.

            Returns:
                self: a instância da classe CreateTable.
        """        
        if not TableUtils.table_exists(self.layer, self.table_name) or self.replace:                 
            ## generate final create table template
            self.template() 
            Utils.spark_instance().sql(self.parsed_template)                                            
            # mark table has been created or replaced by class
            self.has_created_table = True    
        else:
            log(__name__).warning(f"The table already exists, so nothing will be done.")
        return self

class CreateTableBronze(CreateTable):
    """
        Classe que implementa o create table para a camada bronze, é o processo mais simples de todos.

        Args:      
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or {}): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é {}.
            partition_fields (list): lista de campos utilizados como partição da tabela.
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
    """

    layer = "bronze"

    def __init__(self, schema, table_name, yml={}, partition_fields=[], tbl_properties={}, replace=False):        
        super().__init__(schema, table_name, yml, partition_fields, tbl_properties, replace)
    
    def __load_and_verify_params(self):
        if not "dataCarga" in self.schema.names:
            log(__name__).error("Any bronze table must have a dataCarga field.")
            raise Exception("Any bronze table must have a dataCarga field.")

    def parse_fields(self): 
        """
            Método que faz o parse dos campos para o create table na bronze. O método faz a leitura de cada field e gera o seguinte padrão:
            
             .. code-block:: SQL

                field_name field_type nullable or not           
            
            Returns:
                self: a instância da classe CreateTable.
        """
        self.__load_and_verify_params()      
        list_of_parsed_fields = ["{} {}{}".format(field.name, field.dataType.simpleString(), " NOT NULL" if not field.nullable else "") for field in self.schema]
        self.parsed_fields = ", ".join(list_of_parsed_fields)
        return self
    
class CreateTableSilver(CreateTable):
    """
        Classe que implementa um create table para a camada silver, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or {}): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é {}.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "silver".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou {}): instância do objeto yml, é {} caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padrão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "silver"
    
    def __init__(self, 
                 schema, 
                 table_name, 
                 yml={}, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 tbl_properties={},
                 replace=False):        
        super().__init__(schema, table_name, yml, partition_fields, tbl_properties, replace)
        self.fields = self.yml.get("fields", fields)
        self.field_comments = {}
        self.tbl_comment = TableComment(self.layer, self.table_name, self.yml, anonimized_fields, comment, comment_squad)
        self.parsed_fields = ""

    def __load_and_verify_params(self):
        """
            Método interno que executa a verificação dos parâmetros necessários para o create table ser executado.

            Raises:
                ValueError: Caso os parâmetro fields não seja informado; Se nem os campos do dicionário fields possuem um comentário; Se a tabela não tem o campo dataCarga.
        """     
        # comments are required in silver layer
        if len(self.fields) == 0:            
            log(__name__).error(f"If not informed, the fields parameter is required in yml table template for {self.layer} layer.")
            raise ValueError(f"If not informed, the fields parameter is required in yml table template for {self.layer} layer.")                          

        # we manually insert dataCarga to the object (if it is passed it will be generated in comments)
        if not "dataCarga" in self.schema.names:
            log(__name__).error("Any silver table must have a dataCarga field.")
            raise ValueError("Any silver table must have a dataCarga field.")
        else:            
            self.fields["dataCarga"] = ["Data de carga dos dados."]

        # verify if checksum is added to the silver table
        if "checksum" in self.schema.names:            
            self.fields["checksum"] = ["CheckSum/Hash das informações do registro (linha)."]

        # populate comments structure
        for field_name, field_params in self.fields.items():                                    
            self.field_comments[field_name] = field_params[0]

    def parse_fields(self): 
        """
            Método que faz o parse dos campos para o create table na silver. O método faz a leitura de cada field e gera o seguinte padrão:
            
             .. code-block:: SQL
            
                field_name field_type COMMENT 'field_comment' nullable or not         
            
            Returns:
                self: a instância da classe CreateTable.
        """     
        self.__load_and_verify_params()
        list_of_parsed_fields = ["{} {} COMMENT '{}'{}".format(field.name, 
                                                                field.dataType.simpleString(), 
                                                                TableUtils.format_comment_content(self.field_comments[field.name]), 
                                                                " NOT NULL" if not field.nullable else "") for field in self.schema]
        self.parsed_fields = ", ".join(list_of_parsed_fields)
        return self
        
    def execute(self):    
        """
            Método que executa o create table na camada silver e faz a chamada do comment on table, respectivamente. O comentário é gerado novamente, caso a tabela seja criada ou recriada.

            Returns:
                self: a instância da classe CreateTable.
        """      
        super().execute()
        # it forces comment on table when table is recreated
        if self.has_created_table:
            self.tbl_comment.execute()
        return self
    
class CreateTableGold(CreateTableSilver):
    """
        Classe que implementa um create table para a camada gold, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or {}): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é {}.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "gold".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou {}): instância do objeto yml, é {} caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "gold"

    def __init__(self, 
                 schema, 
                 table_name, 
                 yml={}, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 tbl_properties={},
                 replace=False):  
        super().__init__(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, tbl_properties, replace)

class CreateTableDiamond(CreateTableSilver):
    """
        Classe que implementa um create table para a camada diamond, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or {}): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é {}.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "diamond".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou {}): instância do objeto yml, é {} caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "diamond"

    def __init__(self, 
                 schema, 
                 table_name, 
                 yml={}, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 replace=False):  
        super().__init__(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, tbl_properties, replace)

class CreateTableExport(CreateTableSilver):
    """
        Classe que implementa um create table para a camada export, baseado nos padrões do catalogo de dados do Atlan.
        
        Args:
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            yml (yml object, string or {}): caminho do arquivo yml ou instância do objeto com os parâmetros necessários para o Create Table. Padrão é {}.
            partition_fields (list): lista de campos utilizados como partição da tabela.      
            anonimized_fields (list, optional): lista de campos para serem anonimizados.
            comment (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            comment_squad (dict, optional): dicionário com o padrão table_comment_atlan_pattern.key:comment_squad (veja TableComment.table_comment_atlan_pattern). Se não for informado, será inferido a partir do yml.
            fields (dict, optional): dicionário com o padrão field_name:comment. Se não for informado, será inferido a partir do yml.
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            replace (bool, optional): opção que força o replace da tabela caso ela exista. Por padrão é False.

        Attributes:
            layer (str): constante com o valor "export".
            schema (df.schema): schema do dataframe, mesmo que df.schema.
            table_name (str): string que representa o nome da tabela.
            table_path (str): caminho da tabela no lakehouse.
            yml (yml object ou {}}): instância do objeto yml, é {} caso nenhum arquivo seja informado.
            partition_fields (list): lista de campos utilizados como partição da tabela.            
            replace (bool): opção que força o replace da tabela caso ela exista. Por padrão é False.
            has_created_table (bool): parâmetro de controle interno, se True indica que a tabela foi criada ou substituída após a chamada do método execute.
            template (str): template da criação da tabela compilado, preenchido após a chamada do método template().
            fields (dict): dicionário com o padrão field_name:comment. 
            tbl_properties (dict, optional): dicionário com a definição chave e valor de propriedades da tabela. Padrão é {}.
            tbl_comment (TableComment): instância da classe TableComment para gerar o comentário da tabela.
            parsed_fields (str): string com os comentários das fields parseados no padão do create table, preenchido após a chamada do método parse_fields().
    """

    layer = "export"

    def __init__(self, 
                 schema, 
                 table_name, 
                 yml={}, 
                 partition_fields=[],
                 anonimized_fields=[], 
                 comment={}, 
                 comment_squad={}, 
                 fields={},
                 replace=False):  
        super().__init__(schema, table_name, yml, partition_fields, anonimized_fields, comment, comment_squad, fields, tbl_properties, replace)



from dnalib.utils import TableUtils, Utils
from dnalib.log import *

class DDL:    
    """ """    
    def __init__(self, layer):
        self.layer = layer.strip().lower()    

class TableDDL(DDL):
    """ """
    def __init__(self, layer, table_name):
        super().__init__(layer)          
        self.table_name = table_name.strip().lower()

    def describe(self):
        raise NotImplementedError("Method describe() must be implemented.")    

    def create_table(self):
        raise NotImplementedError("Method create_table() must be implemented.")    

    def create_view(self):
        raise NotImplementedError("Method create_view() must be implemented.")    
        
    def drop_table(self):
        raise NotImplementedError("Method drop_table() must be implemented.")        

    def drop_view(self):
        raise NotImplementedError("Method drop_view() must be implemented.")      

    def table_exists(self):
        raise NotImplementedError("Method table_exists() must be implemented.")        

    def view_exists(self):
        raise NotImplementedError("Method view_exists() must be implemented.")        



from pyspark.sql import SparkSession
from .utils import Utils 
from dnalib.log import log
from delta.tables import DeltaTable
from pyspark.sql.functions import monotonically_increasing_id
import json

class TableUtils:

    #  
    #   Description:
    #       This method formats a comment (for field or table).
    #
    #   Parameters:      
    #       comment_content = the content of the comment.
    # 
    @staticmethod
    def format_comment_content(comment_content):
        # remove whitespaces        
        if len(comment_content) > 0:
            value = comment_content.strip()
            # insert ponctuation at the end of string
            if value[-1] != '.':
                value += '.'
            # first digit in upper case        
            return value[0].upper() + value[1:]
        else: 
            return comment_content

    #  
    #   Description:
    #       This method formats a field to a hash type.
    #
    #   Parameters:      
    #       field = field name to be a hashed field
    # 
    @staticmethod
    def hash_field_str(field):        
        return f"sha2(cast({field} as string), 256) AS {field}"

    #  
    #   Description:
    #       This method loads a dataframe from path (usefull if table does not exists in catalog yet).
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    # 
    @staticmethod
    def load_df_from_lakehouse(layer, table_name):        
        root_path = Utils.lakehouse_path()
        try:
           df = Utils.spark_instance().read.load(f'{root_path}/{layer}/{table_name}')
        except:
            log(__name__).error(f"Unable to read '{root_path}/{layer}/{table_name}'")
            raise Exception(f"Unable to read '{root_path}/{layer}/{table_name}'")
        return df
    #  
    #   Description:
    #       This method gets schema table from path (usefull if table does not exists in catalog yet).
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    #     
    @staticmethod
    def table_schema_from_lakehouse(layer, table_name):                        
        return TableUtils.load_df_from_lakehouse(layer, table_name).schema
    
    #  
    #   Description:
    #       This method checks if a view exists from a table name in catalog.
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       view_name = the name of the table.    
    # 
    @staticmethod
    def view_exists(layer, view_name):                                
        return Utils.spark_instance().catalog.tableExists(f"{layer}.{view_name}")        
    
    #  
    #   Description:
    #       This method checks if a table exists in catalog.
    #
    #   Parameters:      
    #       layer = table layer (bronze, silver, gold, diamond or export)
    #       table_name = the name of the table.    
    # 
    @staticmethod
    def table_exists(layer, table_name):                
        return Utils.spark_instance().catalog.tableExists(f"{layer}.{table_name}")
        #try:
            #DeltaTable.forName(Utils.spark_instance(), f"{layer}.{table_name}")
            #return True
        #except:
            #return False
    
    @staticmethod
    def describe_table(layer, table_name):
        df_describe = Utils.spark_instance().sql(f"DESCRIBE EXTENDED {layer}.{table_name}")
        return df_describe.withColumn("rowId", monotonically_increasing_id())

    @staticmethod
    def drop_table(layer, table_name):
        """           
            Esse método executa um drop deep, removendo a tabela e a sua pasta no lakehouse.
    
            Args:
                layer (str): nome da camada.
                table_name (str): nome da tabela para ser dropada.
        """
        Utils.spark_instance().sql(f"drop table if exists {layer}.{table_name}")
        is_removed = Utils.remove_table_from_lakehouse(layer, table_name)
        if not is_removed:
            log(__name__).warning(f"Table {layer}.{table_name} does not exists in lakehouse")  
        return is_removed
    
    @staticmethod    
    def drop_view(layer, view_name):
        """            
            Esse método dropa uma view de maneira safe, isto é caso ela exista.

            Args:
                layer (str): nome da camada.
                view_name (str): nome da view para ser dropada.
        """
        Utils.spark_instance().sql(f"drop view if exists {layer}.{view_name}")     

    @staticmethod
    def table_comment(layer, table_name):   
        if not TableUtils.table_exists(layer, table_name):
            log(__name__).error(f"Table {layer}.{table_name} does not exist")
            raise ValueError(f"Table {layer}.{table_name} does not exist")        
        table_comment = ""    
        df_comment = Utils.spark_instance().sql(f"DESCRIBE EXTENDED {layer}.{table_name}").filter("col_name == 'Comment'").select('data_type')
        if df_comment.count() > 0:                
            table_comment = df_comment.collect()[0][0]
        return table_comment
    
    @staticmethod
    def table_fields_metadata(layer, table_name):
        fields = {}
        if not TableUtils.table_exists(layer, table_name):
            log(__name__).error(f"Table {layer}.{table_name} does not exist")
            raise ValueError(f"Table {layer}.{table_name} does not exist")
        try:
            df_metadata = Utils.spark_instance().sql(f"desc table extended {layer}.{table_name}").withColumn("rowId", monotonically_increasing_id())
            rowId = df_metadata.filter("col_name like '#%'").select('rowId').first()[0]                        
            fields_params = df_metadata.filter(f"(rowId < {rowId}) and col_name != ''").drop('rowId', 'data_type').toJSON().collect()
        except:
            log(__name__).error(f"Unable to load fields from table {layer}.{table_name}.")
            raise ValueError(f"Unable to load fields from table {layer}.{table_name}.") 
        for field_param in fields_params:
            field_param_json = json.loads(field_param)
            fields[field_param_json.get('col_name')] = [field_param_json.get('comment')]
        return fields

    @staticmethod
    def clone_table(source_layer, table_name, target_layer, target_table_name):
        """ 
            Método que executa um clone "safe" entre duas tabelas.

            Args:
                source_layer (str): nome da camada de origem.
                table_name (str): nome da tabela de origem.
                target_layer (str): nome da camada de destino.
                target_table_name (str): nome da tabela de destino.
        """
        if TableUtils.table_exists(f"{source_layer}", table_name):        
            Utils.spark_instance().sql(f"""
                CREATE OR REPLACE TABLE {target_layer}.{target_table_name}
                CLONE {source_layer}.{table_name} 
              """)
            df_comment = Utils.spark_instance().sql(f"DESCRIBE EXTENDED {source_layer}.{table_name}").filter("col_name == 'Comment'").select('data_type')
            if df_comment.count() > 0:                
                table_comment = df_comment.collect()[0][0]
                Utils.spark_instance().sql(f"COMMENT ON TABLE {target_layer}.{target_table_name} IS '{table_comment}'")
        else:
            log(__name__).warning(f"Table {source_layer}.{table_name} does not exists in lakehouse")

    @staticmethod
    def grant_select(layer, table_name, users):
        """ 
            Método que executa grant select para uma tabela e uma lista de usuários.

            Args:
                layer (str): nome da camada.
                table_name (str): nome da tabela.
                users (str ou list): nome ou lista de usuários.
        """
        if isintance(users, str):
            users = [users]
        if TableUtils.table_exists(f"{layer}", table_name):        
            for user in users:                
                try:
                    Utils.spark_instance().sql(f"""GRANT SELECT ON TABLE {layer}.{table_name} TO `{user}`""")           
                except:
                    log(__name__).warning(f"Could not grant SELECT in table {layer}.{table_name} to user {end_user}.")
        else:
            log(__name__).warning(f"Table {layer}.{table_name} does not exists in lakehouse.")
                            


from dnalib.log import log
from .utils import Utils

class FileUtils:

    @staticmethod
    def load_files_dataframe(file_path, ascending=False):
        files_dataframe = None       
        try:            
            files_dataframe = (
                Utils.spark_instance().createDataFrame(Utils.dbutils().fs.ls(file_path))                
                .select('path', 'name', 'modificationTime')
                .orderBy("modificationTime", ascending=ascending)
                .select('path', 'name')                
            )
        except Exception as e:
            log(__name__).error(f"Unable to load files for path {file_path}: {e}")            
            raise Exception(f"Unable to load files for path {file_path}")

        return files_dataframe

    @staticmethod
    def last_file_by_extension(file_path, extension):
        files_dataframe = FileUtils.load_files_dataframe(file_path)
        files_dataframe = files_dataframe.filter(f"lower(name) like '%{extension}'")
        return files_dataframe.first()["path"]

    @staticmethod
    def export_dataframe(df, file_name, **kwargs):
        # use tmp file path to export dataframe
        tmp_file_path = f'/tmp/{file_name}'
        df.repartition(1).write.mode('overwrite').save(tmp_file_path, **kwargs)
        # determine extension based on filename
        file_extension = file_name[file_name.rfind('.'):]
        # get the file name generated by the export
        file_path = FileUtils.last_file_by_extension(f"{tmp_file_path}/", file_extension)
        return file_path






from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text(encoding='UTF-8')

setup(
    name='dnalib',
    version='0.1.19',    
    description='A Python library for fast development of Data Engineering ETL using spark.',    
    author='Charles Gobber',
    author_email='charles26f@gmail.com',
    license='Apache-2',
    packages=find_packages(),
    install_requires=['pyyaml==6.0.2', 'firebase-admin==6.5.0', 'pysftp==0.2.9'],
    long_description=long_description,
    long_description_content_type="text/markdown",
)


from dnalib.utils import Utils, LandingZoneUtils, TableUtils
from dnalib.custom import YamlLoader
from dnalib.log import *
from dnalib.core import CreateTableBronze
from dnalib.writer import WriteModes
from .tables import LayerTable
from enum import Enum
from pyspark.sql.functions import from_utc_timestamp, current_timestamp, concat, col, row_number, when, to_json
from delta.tables import DeltaTable
from pyspark.sql.window import Window
from firebase_admin import credentials, firestore, initialize_app
from tempfile import NamedTemporaryFile
import yaml
import json

class SourceType(Enum):
    """ 
        Enum que define os métodos validos para origem de dados: TXT, CSV, PARQUET, KAFKA e EVENTHUB (WIP); SQLSERVER, ORACLE (implementados).
    """
    TXT = "txt"
    CSV = "csv"
    PARQUET = "parquet"    
    KAFKA = "kafka"
    EVENTHUB = "eventhub"
    SQLSERVER = "sqlserver"
    ORACLE = "oracle"
    FIREBASE = "firebase"

class LandingLoader(Enum):
    """ 
        Enum que define a forma como os dados serão carregados a partir da origem: 
        INCREMENTAL: considera que a origem é um arquivo incremental e define automaticamente que será feito um upsert.
        FULL: considera que a origem é um arquivo completo, e define automaticamente que será feito um overwrite.
        CDC: considera que a origem é um arquivo CDC (funciona somente para SourceType.SQLSERVER) e define automaticamente que será feito um upsert.
        STREAMING: WIP.
    """
    INCREMENTAL = "incremental" # read a single part, and merge data
    FULL = "full" # read all, then overwrite    
    CDC = "cdc" # only for cdc loader    
    STREAMING = "streaming" # for streaming purpose

class Loader:
    """
        A classe Loader é uma implementação de alto nível que define a carga de um arquivo da landing zone (todas ingestões batch são baseadas em arquivos
        seja csv, txt ou banco, que nesse último caso usa parquet).

        Args:
            source_path (str): caminho na landingzone de onde o arquivo será lido.
            table_name (str): string que representa o nome da tabela.          
            loader_type (LandingLoader): algum dos tipos válidos da LandingLoader.            
            name_filter (str, optional): utilizado caso seja necessário filtrar arquivos dentro de um mesmo diretório da landing.

        Attributes:
            source_path (str): caminho na landingzone de onde o arquivo será lido.
            table_name (str): string que representa o nome da tabela.    
            loader_type (LandingLoader): algum dos tipos válidos da LandingLoader.            
            name_filter (str, optional): utilizado caso seja necessário filtrar arquivos dentro de um mesmo diretório da landing.

    """
    def __init__(self, source_path, loader_type):
        self.source_path = source_path        
        self.loader_type = loader_type                
        self.df_loader = None

    def load_incremental(self):
        # TODO: implementar o load do tipo incremental. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_incremental() must be implemented.")

    def load_full(self):
        # TODO: implementar o load do tipo full. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_full() must be implemented.")

    def load_cdc(self):
        # TODO: implementar o load do tipo cdc. (deve retornar um DataFrame Spark)
        raise NotImplementedError("Method load_cdc() must be implemented.")    

    def load(self):
        if self.loader_type.value == LandingLoader.INCREMENTAL.value:
            self.df_loader = self.load_incremental()
        elif self.loader_type.value == LandingLoader.FULL.value:
            self.df_loader = self.load_full()
        elif self.loader_type.value == LandingLoader.CDC.value:
            self.df_loader = self.load_cdc()
        else:
            log(__name__).error(f"Invalid load type {self.loader_type}")
            raise ValueError(f"Invalid load type {self.loader_type}")
        log(__name__).info(f"Load of {self.loader_type.value} type completed successfully.")
        return self

#######################################
#  Firebase Based Loader
######################################

class FireBaseApp: 
    """
        A classe FireBaseApp implementa um singleton para manter uma instancia única de execução do cliente do Firestore.
    """
    _instance = None

    def __init__(self, service_account_data):
        self.service_account_data = service_account_data
        self.db = self.authenticate_db()
        
    def authenticate_db(self):
        with NamedTemporaryFile('w+') as fp:
            fp.write(self.service_account_data)
            fp.seek(0)
            cred = credentials.Certificate(fp.name)
            initialize_app(cred)
            return firestore.client()
        return None
    
    @classmethod
    def instance(cls, service_account_data):
        if cls._instance is None:
            cls._instance = cls(service_account_data)
        return cls._instance

class FireBaseLoader(Loader):
    """
       A classe FireBaseLoader encapsula o processo de carregamento de dados do Firestore para um dataframe Spark. 
    """
    def __init__(self, 
                 source_path,       
                 loader_type,           
                 collection_name, 
                 order_by='__name__', 
                 page_size=1000,
                 init_load=None,
                 end_load=None,
                 flat_json=True):
        super().__init__(source_path, loader_type) 
        self.collection_name = collection_name
        self.order_by = order_by
        self.page_size = page_size
        self.init_load = init_load
        self.end_load = end_load
        self.flat_json = flat_json
        self.service_account_data = Utils.key_vault_value(source_path)   
        self.db = FireBaseApp.instance(self.service_account_data).db
        self.collection = self.load_collection()
        self.docs = []
        self.df_raw = None            

    def load_collection(self):
        """  """
        collection = self.db.collection(self.collection_name)
        if self.init_load:
            filter_init = firestore.FieldFilter(self.order_by, ">=", self.init_load)
            collection = collection.where(filter=filter_init)
        if self.end_load:
            filter_end = firestore.FieldFilter(self.order_by, "<=", self.end_load)
            collection = collection.where(filter=filter_end)
        return collection

    def iterate_documents(self):
        """  """
        cursor = None
        count = 0
        while True:
            docs = []  
            if cursor:
                docs = [snapshot for snapshot in self.collection.limit(self.page_size).order_by(self.order_by).start_after(cursor).stream()]
            else:
                docs = [snapshot for snapshot in self.collection.limit(self.page_size).order_by(self.order_by).stream()]
            count = count + len(docs)
            for doc in docs:
                # return doc to iterate  
                yield doc                
            print(f"Firebase loaded {count} documents.", end="\r")            
            if len(docs) == self.page_size:
                cursor = docs[self.page_size-1]
                continue
            break
    
    def load_incremental(self):
        """  """       
        for doc in self.iterate_documents():            
            self.docs.append(json.dumps(doc.to_dict()))
        # converts all documents to json string, make primitives string to avoid break schema
        if self.flat_json:
            self.df_raw = Utils.spark_instance().read.json(Utils.spark_instance().sparkContext.parallelize(self.docs), primitivesAsString=True)                  
            for field in self.df_raw.schema:
                if field.dataType.simpleString() != "string":
                    self.df_raw = self.df_raw.withColumn(field.name, to_json(col(field.name)))   
        else:
            self.df_raw =  Utils.spark_instance().createDataFrame([[doc] for doc in self.docs], ["json_document"])             
        return self.df_raw
    
    def load_full(self):
        """  """
        return self.load_incremental()

#######################################
#  SQL Based Loaders
######################################

class SqlParquetLandingLoader(Loader):
    """ 
        This implementation is based on our adf process:
            table/date/ -> incremental path
            table/date/ -> cdc path
            table/date/ -> full path = if process run in incremental mode, but you specified a full query
            table/ -> full path
    """
    def __init__(self, source_path, loader_type, table_config):
        super().__init__(source_path, loader_type) 
        self.spark = Utils.spark_instance()  # Obtém a instância do Spark usando Utils.spark_instance()                        
        self.table_config = table_config

    def load_df_parquet(self):
        # le o path mais recente dentro da pasta indicada por source_path
        file_path = LandingZoneUtils.load_last_updated_path(self.source_path)
        # Lê os dados do parquet atual
        try:
            df = self.spark.read.parquet(file_path) 
        except Exception as e:
            log(__name__).error(f"Error reading parquet file: {e}")          
            raise Exception(f"Error reading parquet file: {e}")
        return df

    def load_incremental(self):       
        return self.load_df_parquet()

    def load_full(self):
        return self.load_df_parquet()
        
class SqlServerParquetLandingLoader(SqlParquetLandingLoader):

    # only applied in cdc purposes
    cdc_metadata_columns = ["LAST_UPDATE", "start_lsn", "end_lsn", "seqval", "operation", "update_mask", "command_id", "data_alteracao"]

    def __init__(self, source_path, loader_type, table_config):
        super().__init__(source_path, loader_type, table_config)

    def load_cdc(self):
        df = self.load_incremental()                

        # query que pega a última alteração de cada registro
        df = (df.withColumnRenamed("__$start_lsn", "start_lsn") 
               .withColumnRenamed("__$end_lsn", "end_lsn") 
               .withColumnRenamed("__$seqval", "seqval") 
               .withColumnRenamed("__$operation", "operation") 
               .withColumnRenamed("__$update_mask", "update_mask") 
               .withColumnRenamed("__$command_id", "command_id"))
        
        # view para filtrar as alterações
        df.createOrReplaceTempView("vw_table_logs")
        self.spark.sql(f"""
            CREATE OR REPLACE TEMP VIEW vw_table_logs_rownumber AS (
                SELECT * FROM (
                    SELECT 
                        *, 
                        ROW_NUMBER() OVER(PARTITION BY {self.table_config['key']} ORDER BY data_alteracao DESC, command_id DESC) AS LAST_UPDATE FROM vw_table_logs
                    WHERE operation <> 3                                                
                )
                WHERE LAST_UPDATE = 1
            );
        """)

        # filtra as operações de insert e update
        df_upinsert = self.spark.sql(""" 
            SELECT * 
            FROM vw_table_logs_rownumber 
            WHERE operation IN (2, 3, 4);
            """).drop(*self.cdc_metadata_columns)                      
        return df_upinsert        

class OracleParquetLandingLoader(SqlParquetLandingLoader):

    oracle_metadata_columns = ["Dt_Inclusao"]

    def __init__(self, source_path, loader_type, table_config):
        super().__init__(source_path, loader_type, table_config)

    def load_incremental(self):     
        df_upinsert = super().load_incremental()          
        return df_upinsert.drop(*self.oracle_metadata_columns)

    def load_full(self):
        df_overwrite = super().load_full()
        return df_overwrite.drop(*self.oracle_metadata_columns)

class LoaderFactory:
    @staticmethod
    def create_loader(source_type, 
                      source_path, 
                      loader_type, 
                      table_config, 
                      collection_name=None,                      
                      order_by='__name__', 
                      page_size=1000, 
                      init_load=None, 
                      end_load=None,
                      flat_json=True):
        if source_type.value == SourceType.SQLSERVER.value:
            return SqlServerParquetLandingLoader(source_path, loader_type, table_config)
        elif source_type.value == SourceType.ORACLE.value:
            return OracleParquetLandingLoader(source_path, loader_type, table_config)
        elif source_type.value == SourceType.FIREBASE.value:
            return FireBaseLoader(source_path, loader_type, collection_name, order_by, page_size, init_load, end_load, flat_json)
        else:
            log(__name__).error(f"file_loader could not be created for source_type {source_type}")
            raise ValueError(f"file_loader could not be created for source_type {source_type}")
        
class BronzeTable(LayerTable):

    layer = "bronze"

    def __init__(self,                  
                 table_name,                 
                 source_path,
                 yaml_config_path,
                 table_sufix = "",
                 source_type=SourceType.SQLSERVER, 
                 loader_type=LandingLoader.FULL,
                 source_df=None, 
                 name_filter="", 
                 unzip=True, 
                 include_checksum=False,  
                 collection_name=None,               
                 order_by = '__name__',
                 page_size = 1000,
                 init_load = None, 
                 end_load = None,
                 flat_json=True):
        super().__init__(self.layer, f"{table_sufix}_{table_name}", source_df, include_checksum)        
        self.source_path = source_path
        self.name_filter = name_filter
        self.table_sufix = table_sufix
        self.source_type = source_type
        self.yaml_config_path = yaml_config_path
        self.table_config = Utils.yaml_table_parameters(table_name=f"{table_sufix}_{table_name}", yml_file_path=yaml_config_path)
        self.loader_type = loader_type
        self.unzip = unzip      
        self.collection_name = collection_name     
        self.order_by = order_by
        self.page_size = page_size
        self.init_load = init_load
        self.end_load = end_load  
        self.flat_json = flat_json
        self.except_fields = {}

        # Usa a fábrica para criar o file_loader correto
        self.file_loader = LoaderFactory.create_loader(            
            self.source_type,
            self.source_path,
            self.loader_type,            
            self.table_config,     
            self.collection_name,       
            self.order_by,
            self.page_size,
            self.init_load,
            self.end_load, 
            self.flat_json
        )        

    def parse_df_source(self):
        """
            Método que executa o carregamento de dados da camada landing para a bronze

            Returns:
                source_df (spark DataFrame): dataframe carregado a partir da camada source, caso source_df seja None.

        """         
        # load strategy in this case is to check if the table exists        
        return self.file_loader.load().df_loader

    def load_fields_from_source(self):                
        return "*"
    
    def create_table(self, yml={}, partition_fields=[], tbl_properties={}, replace=False):
        if not self.has_loaded:
            self.load()
            log(__name__).warning(f"The load() method will be called internally because you call create_table first.") 
        self.creat_tbl = CreateTableBronze(self.target_df.schema, self.table_name, yml, partition_fields, tbl_properties, replace).execute()
        return self
    
    # implementar lógica usando mode = None (nesse caso a decisão usa o LandingLoader como parâmetro caso contrário fica a critério de quem escolheu)
    def persist(self, mode=WriteModes.OVERWRITE, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None):        
        has_delete = False                
        # when you have cdc ingestion, you may need to upsert
        if self.loader_type.value == LandingLoader.CDC.value:
            mode = WriteModes.UPSERT
            merge_condition = self.table_config["merge"]
            has_delete = True        
        elif self.loader_type.value == LandingLoader.INCREMENTAL.value:
            mode = WriteModes.UPSERT
            merge_condition = self.table_config["merge"]        

        # calling super persist method
        super().persist(mode=mode, partition_fields=partition_fields, optimize=optimize, source_df_name=source_df_name, update_df_name=update_df_name, merge_condition=merge_condition)

        # only for cdc purposes
        if has_delete:            
            key_columns = ",".join(self.table_config["key"]) if isinstance(self.table_config["key"], list) else self.table_config["key"]
            delete_condition = f"SELECT concat({key_columns}) FROM vw_table_logs_rownumber WHERE operation = 1"
            self.writer.delete(key_columns, delete_condition)       
        return self     




from dnalib.log import log
from dnalib.utils import Utils
from delta.tables import DeltaTable
from enum import Enum

class WriteType(Enum):
    """ Classe com os tipos de escrita, batch e streaming """
    BATCH = "batch"
    STREAMING = "streaming"

class WriteModes(Enum):
    """ Classe com os modos de escritas válidos: overwrite, append e upsert. """
    OVERWRITE = "overwrite"    
    APPEND = "append"    
    UPSERT = "upsert"    

class TableWriter:
    """ Interface """

    def __init__(self, table_name, layer):
        self.table_name = table_name
        self.layer = layer        

    def append(self):
        # TODO: implementar a escrita via append. 
        raise NotImplementedError("Method append() must be implemented.")

    def overwrite(self):
        # TODO: implementar a escrita via overwrite. 
        raise NotImplementedError("Method overwrite() must be implemented.")

    def upsert(self):
        # TODO: implementar a escrita via upsert. 
        raise NotImplementedError("Method upsert() must be implemented.")

    def optimize(self):        
        """ 
            Método interno para executar o OPTIMIZE.
        """     
        log(__name__).info(f"Optimizing table: {self.layer}.{self.table_name}.") 
        Utils.spark_instance().sql(f"OPTIMIZE {self.layer}.{self.table_name}")

class BatchTableWriter(TableWriter):

    def __init__(self, table_name, layer):
        super().__init__(table_name, layer)

    def __overwrite_or_append(self, df, mode, partition_fields=[]):
        # load df and mode
        df_write = df.write.format("delta").mode(mode.value)        
        # verify partitions
        if len(partition_fields) > 0:
            df_write = df_write.partitionBy(partition_fields)          
        # saving data in delta table                                
        df_write.saveAsTable(f"{self.layer}.{self.table_name}")  

    def upsert(self, df_update, merge_condition, source_df_name="source", update_df_name="update", has_checksum_field=False):
        """
            Método interno que executa o upsert.
                
            Args:   
                merge_condition (str): uma condição valida para fazer merge entre o "source_df_name" e o "update_df_name".
                has_checksum_field (bool): uma flag que indica se a tabela tem ou não o campo de checksum.            
        """
        df_source = DeltaTable.forName(Utils.spark_instance(), f"{self.layer}.{self.table_name}")
        # it improves merge performance in write operations
        update_condition = None
        if has_checksum_field:
            update_condition = f"nvl({source_df_name}.checksum, '') != {update_df_name}.checksum"
        # runing a normal merge operation
        (df_source.alias(source_df_name)
            .merge(
                df_update.alias(update_df_name),
                merge_condition
            )
            .whenMatchedUpdateAll(condition=update_condition)
            .whenNotMatchedInsertAll()
            .execute()) 
        return self
    
    def delete(self, key_columns, delete_condition):
        """
        Método que executa a exclusão de dados da tabela com base em um conjunto de chaves e uma condição.

        Args:
            key_columns (str): Chaves para realizar o delete.
        
        Returns:
            self: uma instância da classe TableWriter.
        """                    
        try:            
            df_delete = Utils.spark_instance().sql(f"""DELETE FROM {self.layer}.{self.table_name}
                           WHERE concat({key_columns}) IN ({delete_condition}) """)                        
            num_affected_rows = df_delete.first()["num_affected_rows"]
            log(__name__).info(f"{num_affected_rows} records where deleted from {self.layer}.{self.table_name}.")
        except:
            log(__name__).error(f"Error while trying to delete records from {self.layer}.{self.table_name}.")
            raise Exception(f"Error while trying to delete records from {self.layer}.{self.table_name}.")

        return self

    def overwrite(self, df, partition_fields=[]):
        """ """
        self.__overwrite_or_append(df, WriteModes.OVERWRITE, partition_fields)
        return self

    def append(self, df, partition_fields=[]):
        """ """
        self.__overwrite_or_append(df, WriteModes.APPEND, partition_fields)
        return self

    def persist(self, df, mode=WriteModes.OVERWRITE, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None, has_checksum_field=False):
        """ """
        df_count = df.count() 
        if df_count == 0:                                        
            log(__name__).warning(f"No data to persist in {self.table_name}, so nothing will be done.")       
        else:     
             # overwrite or append has same sintax
            if mode.value == WriteModes.OVERWRITE.value:
                self.overwrite(df, partition_fields)
            elif mode.value == WriteModes.APPEND.value:         
                self.append(df, partition_fields)        
            elif mode.value == WriteModes.UPSERT.value:
                if merge_condition is None:
                    log(__name__).error(f"The merge_condition parameter is required for upsert mode")           
                    raise Exception(f"The merge_condition parameter is required for upsert mode")                
                self.upsert(df, merge_condition, source_df_name, update_df_name, has_checksum_field)
            else:
                log(__name__).error(f"The mode parameter {mode} is not valid. Valid values are: {[m.value for m in WriteModes]}")            
                raise Exception(f"The mode parameter {mode} is not valid. Valid values are: {[m.value for m in WriteModes]}")            
            log(__name__).info(f"Sucessfully persisted {df_count} rows to {self.layer}.{self.table_name} using {mode.value} mode.") 
            # runing optimize operation
            if optimize:            
                self.optimize()                
        return self

class StreamingTableWriter(TableWriter):
    """
        Classe para escrita de streaming
    """

    def __init__(self, table_name, layer, checkpoint_path="_checkpoints", processing_time="10 minutes"):
        super().__init__(table_name, layer)
        self.checkpoint_path = checkpoint_path
        self.processing_time = processing_time
        self.checkpoint_location = Utils.lakehouse_path(f"{self.layer}/{self.checkpoint_path}/{self.table_name}")    

    def upsert(self, df, merge_condition, source_df_name="source", update_df_name="update", drop_keys=[]):
        layer = self.layer
        table_name = self.table_name
        # inner method to execute merge based on our implementation
        def merge_micro_df(micro_df, batch_id):
            if not microdf.isEmpty():                
                df_source = DeltaTable.forName(Utils.spark_instance(), f"{layer}.{table_name}")
                ## Merge
                df_updates = microdf.dropDuplicates(drop_keys)
                (df_source.alias(source_df_name)
                    .merge(
                        df_update.alias(update_df_name),
                        merge_condition
                    )
                    .whenMatchedUpdateAll()
                    .whenNotMatchedInsertAll()
                    .execute()) 
        # call writestream
        (df.writeStream.format("delta")
            .outputMode('append')  
            .trigger(processingTime=self.processing_time)
            .foreachBatch(merge_micro_df) 
            .option("checkpointLocation", self.checkpoint_location)
            .start())
        return self

    def append(self, df):
        (df.writeStream.format("delta")
            .outputMode('append')
            .trigger(processingTime=self.processing_time)
            .option("checkpointLocation", self.checkpoint_location) 
            .start(Utils.lakehouse_path(f"{self.layer}/{self.table_name}")))
        return self

    def persist(self, df, mode=WriteModes.APPEND, partition_fields=[], optimize=True, source_df_name="source", update_df_name="update", merge_condition=None, drop_keys=[]):
        if mode.value == WriteModes.APPEND.value:
            self.append(df)
        elif mode.value == WriteModes.UPSERT.value:
            if merge_condition is None:
                log(__name__).error(f"The merge_condition parameter is required for upsert mode")           
                raise Exception(f"The merge_condition parameter is required for upsert mode")
            self.upsert(df, merge_condition, source_df_name, update_df_name, drop_keys)
        else:
            log(__name__).error(f"The mode parameter {mode} is not valid. Valid values are: {[m.value for m in WriteModes]}")            
            raise Exception(f"The mode parameter {mode} is not valid. Valid values are: {[m.value for m in WriteModes]}")  

        # implementar logica de desligamento do streaming
        
        return self
    